/* Debug Header */
#if DEBUG
# define IFDEBUG(x) do { if (debug) x; }while(0)
# define IFDEBUG5(x) do { if (debug) x; }while(0)
#else
# define IFDEBUG(x)
# define IFDEBUG5(x)
#endif
/* End Debug Header */


#if HAVE_CONFIG_H
# include <config.h>
#endif

/* This file if for config file operations and configurations */

#include "config.h"

#include <sys/file.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <stdarg.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>

#include "configuration.h"
#include "extern.h"
#include "defines.h"

/* main config file handler
 * parse_config():
 *
 * Accepts: argc, and argv
 * Returns: 1 for error, 0 for sucsess
 *
 * purpose: populate config structure.
 */

int parse_config(int argc, char **argv) {

	char conf_file[]=CONFIG_FILE_NAME;
	int fd, i;
	

	config=(config_struct *)calloc(sizeof(config_struct), 1);
	config->port=(port_struct **)calloc(sizeof(port_struct *), MAX_HTTPPORTS+MAX_SOCKSPORTS+1);
	for(i=0; i<MAX_HTTPPORTS+MAX_SOCKSPORTS+1; i++)
		config->port[i]=(port_struct *)calloc(sizeof(port_struct), 1);
	config->port[0]->port=25;
	config->port[0]->type=SMTP;
	config->ports=1;
	config->children=1; /* this is to ensure we have 1 more than the number of ports+parent */

	config->exitval=(struct exit_struct *)calloc(sizeof(struct exit_struct), 1);
	config->exitval->socks=1; /* set default */
	config->exitval->http=1; /* set default */
	config->exitval->relay=1; /* set default */
	config->exitval->body=1; /* set default */
	config->exitval->rdns=1; /* set default */
	config->exitval->trap=1; /* set default */

	config->file=(struct files_struct *)calloc(sizeof(struct files_struct), 1);

	if (fd=lock_config(conf_file)) {
		
		IFDEBUG(print_debug(4, "Reading config file\n"));
		read_config(fd);

		IFDEBUG(print_debug(4, "Unlocking/closing config file\n"));
		release_config(fd);

	}

	IFDEBUG(print_debug(4, "Reading command line\n"));
	cli_config(argc, argv);
}

void release_config(int fd) {

	IFDEBUG(print_debug(3, "Closing Config File\n"));
	lockf(fd, F_ULOCK, 0);
	close(fd);

}

/* Lock config filei
 * lock_config():
 *
 * accepts: filename
 * returns: file descriptor
 *
 * purpose: open and lock the config file
 */

int lock_config(char *filename) {

	int fd;

	IFDEBUG(print_debug(1, "Opening Config File\n"));
	fd=open(filename, O_RDONLY);
	if (errno) {
		fprintf(stderr, "%s: ", filename);
		perror("");
		return 0;
	}
	
	while (lockf(fd, F_TEST, 0)) {
		if (errno != EACCES) {
			perror("Error Locking Config file");
			exit(1);
		}
		fprintf(stderr, "Waiting for config file lock!\n");
		sleep(5);
	}

	IFDEBUG(print_debug(4, "Locking Config File For exclusive access\n"));
	lockf(fd, F_LOCK, 0);
	IFDEBUG(print_debug(4, "Got Config Lock\n"));

	return fd;
}
/* read config
 * read_config()
 * accepts: open and locked filedescriptor for config file
 * returns: nothing
 *
 * purpose: read in configfile and set options/globals
 */
void read_config(int fd) {
	
	int i;
	char *p, *line;
	char *buf;

	print_debug(4, "Reading Config File into Buffer\n");
	config->config=read_file(fd);

	buf=(char *)calloc(sizeof(char), FILE_BUF);
	line=config->config;
	while (line && *line) {
		p=(char *)strchr(line, '\n');
		if (p) {
			memset(buf, '\0', FILE_BUF);
			strncpy(buf, line, (p-line)<FILE_BUF?(size_t)(p-line):(size_t)FILE_BUF-1);
			if (*buf != '#') parse_line(buf);
			p++;
		}
		line=p;
	}
	free(buf);
}

void parse_line(char *buf) {

	char *p;
	char *tp;
	int i;

	/* Config options:
	 *
	 * ADD_HTTP - Bool: Add to Open HTTP-Proxy list
	 * ADD_RDNS - Bool: Add to RDNS failure list
	 * ADD_SMTP - Bool: Add to Open Relay list
	 * ADD_SOCKS - Bool: Add to Open Socks list
	 * ADD_TRAP - Bool: Add to SPAM TRAP list on delivery attempt.
	 * DB_HOME - Use DB's as specified in DBMODE 'local' foe local - hostname for remote
	 * DB_PORT - port of database (ignored for DB_HOME='local')
	 * DB_TYPE - DB type: mysql, local, berkeley etc..
	 * DB_MODE - Bool: DB Useage On/Off
	 * EXIT_LEVEL - (int) Error level to exit with if reject hit (0 = don't exit)
	 * FILE_HTTP - Open HTTP-Proxy DB FIlename
	 * FILE_RDNS - DB Filename for hosts with no RDNS
	 * FILE_SMTP - Open Relay DB Filename
	 * FILE_SOCK - Socks DB Filename
	 * FILE_TRAP - DB File for IP Addresses where SPAMTRAP was triggered
	 * LOCAL_DB - DNS Zone files or Sendmail Access lists
	 * MODE - AS deamon, plugin or standalone
	 * PORTS - comma delimited list of ports to check
	 * REJECT_BODY - Bool: Reject on Other Received header matches.
	 * REJECT_HTTP - Bool: Reject on Open HTTP-Proxy
	 * REJECT_RDNS - Bool: Reject on RDNS failure
	 * REJECT_SMTP - Bool: Reject on Open Relay
	 * REJECT_SOCK - Bool: Reject on Open Socks
	 * REJECT_TRAP - Bool: Reject on SPAM TRAP delivery attempt.
	 * SECURITY - Mode to run in (Relaxed/Normal/Paranoid)
	 * THREAD - Bool: Run single/multi threaded
	 * VERBOSE - Reject with SMTP-Errors or Bit-Bucket (silent accept and drop) (True for SMTP Errors)
	 * TEST_HOST - Multiple: <host>:<port> for proxy testing.
	 */

	while (*buf && (*buf == '\r' || *buf == '\n' || *buf == '\t')) buf++; /* Skip to first char */

	p=buf;
	while (*p) {
		*p=tolower(*p); /* kill case */
		p++;
	}
	p=buf;

	IFDEBUG(print_debug(5, "Buf: \"%s\"\n", buf));
	i=getoption(buf);
	IFDEBUG(print_debug(5, "getoption() returned: %i\n", i));
	if (!i) return;
	buf=(char *)strchr(buf, '=');
	buf++;
	switch(i) {
		case 1:
			config->add_http=getbool(buf);
			break;
		case 2:
			config->add_rdns=getbool(buf);
			break;
		case 3:
			config->add_smtp=getbool(buf);
			break;
		case 4:
			config->add_sock=getbool(buf);
			break;
		case 5:
			config->add_trap=getbool(buf);
			break;
		case 6:
			config->dbhome=(char *)calloc(sizeof(char), strlen(buf)+1);
			strcpy(config->dbhome, buf);
			break;
		case 7:
			config->dbport=atoi(buf);
			break;
		case 8:
			config->dbmode=getmode(buf);
			break;
		case 9:
			config->dbtype=gettype(buf);
			break;
		case 10:
			config->exitval->socks=atoi(buf);
			break;
		case 11:
			config->file->http=(char *)calloc(sizeof(char), strlen(buf)+1);
			strcpy(config->file->http, buf);
			break;
		case 12:
			config->file->rdns=(char *)calloc(sizeof(char), strlen(buf)+1);
			strcpy(config->file->rdns, buf);
			break;
		case 13:
			config->file->smtp=(char *)calloc(sizeof(char), strlen(buf)+1);
			strcpy(config->file->smtp, buf);
			break;
		case 14:
			config->file->sock=(char *)calloc(sizeof(char), strlen(buf)+1);
			strcpy(config->file->sock, buf);
			break;
		case 15:
			config->file->trap=(char *)calloc(sizeof(char), strlen(buf)+1);
			strcpy(config->file->trap, buf);
			break;
		case 16:
			config->file->localdb=(char *)calloc(sizeof(char), strlen(buf)+1);
			strcpy(config->file->localdb, buf);
			break;
		case 17:
			config->mode=getmode(buf);
			break;
		case 18:
			p=buf;
			for (i=0; i<MAX_SOCKSPORTS; i++) {
				if (!*p) break;
				while (*p && *p != '\t' && *p != ',' && *p != '\r' && *p != '\n') p++;
				*p='\0';
				config->port[config->ports]->port=atoi(buf);
				config->port[config->ports]->type=SOCKS;
				config->children++;
				config->ports++;
				p++;
				while (*p && (*p == ' ' || *p == '\t' || *p == ',' || *p == '\r' || *p == '\n')) p++;
				buf=p;
				if (!*buf) break;
			}
			break;
		case 19:
			p=buf;
			for (i=0; i<MAX_HTTPPORTS; i++) {
				if (!*p) break;
				while (*p && *p != '\t' && *p != ',' && *p != '\r' && *p != '\n') p++;
				*p='\0';
				config->port[config->ports]->port=atoi(buf);
				config->port[config->ports]->type=HTTP;
				config->children++;
				config->ports++;
				p++;
				while (*p && (*p == ' ' || *p == '\t' || *p == ',' || *p == '\r' || *p == '\n')) p++;
				buf=p;
				if (!*buf) break;
			}
			break;
		case 20:
			config->reject_body=getbool(buf);
			break;
		case 21:
			config->reject_http=getbool(buf);
			break;
		case 22:
			config->reject_rdns=getbool(buf);
			break;
		case 23:
			config->reject_smtp=getbool(buf);
			break;
		case 24:
			config->reject_sock=getbool(buf);
			break;
		case 25:
			config->reject_trap=getbool(buf);
			break;
		case 26:
			config->level=getsecurity(buf);
			break;
		case 27:
			config->threads=getbool(buf);
			break;
		case 28:
			config->bitbucket=getbool(buf);
			break;
		case 29:
			debug=atoi(buf);
			break;
		case 30:
			p=buf;
			for (i=0;;i++) {
				char *x;
				x=NULL;
				if (!*p) break;
				while (*p && *p != ' ' && *p != '\t' && *p != ',' && *p != '\r' && *p != '\n') p++;
				*p='\0';
				config->spamtrapuser=(char **)realloc(config->spamtrapuser, sizeof(char *)*(i+2));
				config->spamtrapdom=(char **)realloc(config->spamtrapdom, sizeof(char *)*(i+2));
				config->spamtrapuser[i+1]=NULL;
				config->spamtrapdom[i+1]=NULL;
				config->spamtrapuser[i]=(char *)calloc(sizeof(char), strlen(buf)+1);
				strcpy(config->spamtrapuser[i], buf);
				if (x=strrchr(config->spamtrapuser[i], '@')) {
					*x++='\0';
					config->spamtrapdom[i]=x;
				}
				p++;
				while (*p && (*p == ' ' || *p == '\t' || *p == ',' || *p == '\r' || *p == '\n')) p++;
				buf=p;
				if (!*buf) break;
			}
			break;
		case 31:
			config->exitval->relay=atoi(buf);
			break;
		case 32:
			config->exitval->http=atoi(buf);
			break;
		case 33:
			config->exitval->body=atoi(buf);
			break;
		case 34:
			config->exitval->rdns=atoi(buf);
			break;
		case 35:
			config->exitval->trap=atoi(buf);
			break;
		case 36:
			config->strictsmtp=getbool(buf);
			break;
		case 37:
			config->listenport=atoi(buf);
			break;
		case 38:
			config->listenaddr=(char *)calloc(sizeof(char), strlen(buf)+1);
			strcpy(config->listenaddr, buf);
			break;
		case 39:
			// config->children=atoi(buf);
			break;
		case 40:
			config->file->logfile=(char *)calloc(sizeof(char), strlen(buf)+1);
			strcpy(config->file->logfile, buf);
			break;
		case 41:
			config->file->errfile=(char *)calloc(sizeof(char), strlen(buf)+1);
			strcpy(config->file->errfile, buf);
			break;
		case 42:
			config->targetport=atoi(buf);
			break;
		case 43:
			config->targethost=(char *)calloc(sizeof(char), strlen(buf)+1);
			strcpy(config->targethost, buf);
			break;
		case 44:
			config->lockfile=(char *)calloc(sizeof(char), strlen(buf)+1);
			strcpy(config->lockfile, buf);
			break;
		case 45:
			p=buf;
			for (i=0;;i++) {
				if (!*p) break;
				while (*p && *p != ' ' && *p != '\t' && *p != ',' && *p != '\r' && *p != '\n') p++;
				*p='\0';
				if (tp=(char *)strchr(buf, ':')) {
					config->testhost=(th_struct **)realloc(config->testhost, sizeof(th_struct *)*(i+2));
					config->testhost[i+1]=NULL;
					config->testhost[i]=(th_struct *)calloc(sizeof(th_struct), 1);
					config->testhost[i]->hostname=(char *)calloc(sizeof(char), strlen(buf)+1);
					strncpy(config->testhost[i]->hostname, buf, (tp-buf));
					tp++;
					config->testhost[i]->port=(int)atoi(tp);
				}
				p++;
				while (*p && (*p == ' ' || *p == '\t' || *p == ',' || *p == '\r' || *p == '\n')) p++;
				buf=p;
				if (!*buf) break;
			}
			break;

		default:
			break;
		}
}

/* getbool():
 * accepts: pointer to string
 * returns: 1 (for true) or 0 (for false)
 *
 * purpose: will return 1 (one) when matching '1', 'on', 'yes' or 'true'
 *          returns 0 (zero) in all other cases.
 *
 */

int getbool(char *buf) {

	char *state[] = { "1",
	                  "on",
	                  "true",
	                  "yes",
	                  "" };
	int i;

	if (!buf || !*buf) return 0;

	for(i=0;;i++) {
		if (!*state[i]) goto false;
		if (!strcmp(buf, state[i])) goto true;
	}
false:
	IFDEBUG(print_debug(4, "getbool():false: '%s' returning: 0\n", buf));
	return 0;
true:
	IFDEBUG(print_debug(4, "getbool():true: '%s' returning: 1\n", buf));
	return 1;
}

int gettype(char *buf) {

	/* Types:
	 * 0 - local
	 * 1 - mysql
	 * 2 - berkeley
	 * 3 - oracle
	 */

	char *types[] = { "local",
	                  "mysql",
	                  "berkeley",
	                  "oracle"
	                  "" };
	int i;

	IFDEBUG(print_debug(4, "gettype(): %s\n", buf));
	if (!buf || !*buf) return 0;

	for(i=0;;i++) {
		if (!strlen(types[i])) return 0;
		if (!strcmp(buf, types[i])) return i;
	}

}

int getsecurity(char *buf) {

	/* types:
	 * 0 - paranoid
	 * 1 - normal
	 * 2 - relaxed 
	 */

	char *types[] = { "paranoid",
	                  "normal",
			              "relaxed",
	                  "" };
	int i;

	IFDEBUG(print_debug(4, "getsecurity(): %s\n", buf));
	if (!buf || !*buf) return 0;

	for(i=0;;i++) {
		if (!strlen(types[i])) return 0;
		if (!strcmp(buf, types[i])) return i;
	}
	return 0;
}

int getmode(char *buf) {

	/* types:
	 * 0 - standalone
	 * 1 - daemon
	 * 2 - plugin (Netscape)
	 * 3 - plugin (iMS)
	 */

	char *types[] = { "local",
	                  "daemon",
			              "nms",
			              "ims",
	                  "" };
	int i;

	IFDEBUG(print_debug(4, "getmode(): %s\n", buf));
	if (!buf || !*buf) return 0;

	for(i=0;;i++) {
		if (!strlen(types[i])) return 0;
		if (!strcmp(buf, types[i])) return i;
	}
	return 0;
}

int getoption(char *buf) {

	int i;
	char *p;

	char *options[] = { "add_http",		/* 1 */
											"add_rdns",		/* 2 */
											"add_smtp",		/* 3 */
											"add_socks",		/* 4 */
											"add_spamtrap",		/* 5 */
											"dblisten_addr", /*6 */
											"dblisten_port", /*7 */
											"db_mode",		/* 8 */
											"db_type",		/* 9 */
											"socks_exit_level",		/* 10 */
											"file_http",		/* 11 */
											"file_rdns",		/* 12 */
											"file_smtp",		/* 13 */
											"file_socks",		/* 14 */
											"file_spamtrap",		/* 15 */
											"local_db",		/* 16 */
											"mode",		/* 17 */
											"socks_ports",		/* 18 */
											"http_ports",		/* 19 */
											"reject_body",		/* 20 */
											"reject_http",		/* 21 */
											"reject_rdns",		/* 22 */
											"reject_smtp",		/* 23 */
											"reject_socks",		/* 24 */
											"reject_spamtrap",		/* 25 */
											"security",		/* 26 */
											"threads",		/* 27 */
											"verbose_reject",		/* 28 */
											"debug",		/* 29 */
											"spam_addresses",		/* 30 */
											"relay_exit_level",		/* 31 */
											"http_exit_level",		/* 32 */
											"body_exit_level",		/* 33 */
											"rdns_exit_level",		/* 34 */
											"spamtrap_exit_level",		/* 35 */
											"strict_smtp",		/* 36 */
											"listen_port",		/* 37 */
											"listen_addr",		/* 38 */
											"workers",		/* 39 */
											"logfile",		/* 40 */
											"errfile",		/* 41 */
											"smtp_server_port",		/* 42 */
											"smtp_server_addr",		/* 43 */
											"lockfile",		/* 44 */
											"proxy_testhosts", /* 45 */
											"" };

	if (!buf || !*buf) return 0;
	p=(char *)strchr(buf, '=');
	if (!p) return 0;
	for (i=0;*options[i];i++) {
		if (!strlen(options[i])) return 0;
		// printf("Attempting to match the first %i chars of: %s & %s\n", (int)(p-buf), options[i], buf);
		if (!strncmp(options[i], buf, (size_t)(p-buf))) return i+1; /* Add 1 to the return! */
	}
	return 0;
}

/* close config */


/* read_file()
 * accepts: opend filedescriptor
 * returns: pointer
 *
 * purpose: allocate memory and read-in contents of file
 */
char *read_file(int fd) {

	int l, t;
	char *p, *buf;

	p=NULL;
	t=0;

	buf=(char *)malloc(FILE_BUF*sizeof(char));
	l=1;
	while (l) {
		memset(buf, '\0', FILE_BUF);
		l=read(fd, buf, FILE_BUF-1);
		if (l==-1) {
			perror("read_file()");
			exit (1);
		}
		IFDEBUG(print_debug(5, "Read %i chars into the buffer\n", l));
		if (!l) break;
		t+=l;
		if (p) {
			p=(char *)realloc(p, t*sizeof(char));
		} else {
			p=(char *)calloc(sizeof(char), FILE_BUF);
		}
		strcat(p, buf);
	}
	free(buf);
	return p;
}


int cli_config(int argc, char **argv) {

	int i;

	for ( i=1 ; i<argc; i++ ) {

		if (argv[i][0] = '-' && argv[i+1]) {
			switch (argv[i][1]) {
				case 'h':
					config->hostname=argv[++i];
					break;
				case 'p':
					if (argc<i) usage(argv[0]);
					port=atoi(argv[++i]);
					if (!port) usage(argv[0]);
					break;
				case 'a':
					mode=0;
					break;
				case 'r':
					mode=2;
					break;
				case 'n':
					mode=1;
					break;
				case 'v':
					verbose=1;
					break;
				case 'w':
					dbon=1;
					break;
				case 'd':
					debug=atoi(argv[++i]);
					break;
				default:
					usage(argv[0]);
					break;
			}
		}
	}
}

void print_config(void) {

	int i;

	print_debug(5,"Config File Reads:\n%s\n", config->config);
	print_debug(1,"Add open HTTP Proxies: %s\n", config->add_http?"true":"false");
	print_debug(1,"Add open SMTP Relays: %s\n", config->add_smtp?"true":"false");
	print_debug(1,"Add open Socks Proxies: %s\n", config->add_sock?"true":"false");
	print_debug(1,"Add hosts with no reverse DNS: %s\n", config->add_rdns?"true":"false");
	print_debug(1,"Add hosts that try to use the SPAM trap: %s\n", config->add_trap?"true":"false");
	print_debug(1,"File listing open HTTP Proxies: %s\n", config->file->http);
	print_debug(1,"File listing open SMTP Relays: %s\n", config->file->smtp);
	print_debug(1,"File listing open Socks Proxies: %s\n", config->file->sock);
	print_debug(1,"File listing hosts with no reverse DNS: %s\n", config->file->rdns);
	print_debug(1,"File listing hosts that tried to use the SPAM trap: %s\n", config->file->trap);
	print_debug(1,"Reject open HTTP Proxies: %s\n", config->add_http?"true":"false");
	print_debug(1,"Reject open Socks Proxies: %s\n", config->reject_sock?"true":"false");
	print_debug(1,"Reject open SMTP Relays: %s\n", config->reject_smtp?"true":"false");
	print_debug(1,"Reject hosts with no reverse DNS: %s\n", config->reject_rdns?"true":"false");
	print_debug(1,"Reject mail that is passed through open servers (socks/http proxies/smtp relays): %s\n", config->reject_body?"true":"false");
	print_debug(1,"Reject mailers that have used the spamtrap: %s\n", config->reject_trap?"true":"false");

	print_debug(1,"Database host: %s\n", config->dbhome);
	print_debug(1,"Database port: %i\n", config->dbport);
	print_debug(1,"Database mode: %i\n", config->dbmode);
	print_debug(1,"Database type: %i\n", config->dbtype);

	print_debug(1,"%i Ports to test for open servers: \n", config->ports);
	for (i=0;i<config->ports;i++) {
		print_debug(1, "%i [%s]\n", config->port[i]->port, config->port[i]->type?config->port[i]->type==SMTP?"SMTP":config->port[i]->type==HTTP?"HTTP":"SOCKS":"IDLE");
	}

	if (config->spamtrapuser) {
		for (i=0;config->spamtrapuser[i];i++) {
			print_debug(1, "Spam trap email address(s): %s @ %s\n", config->spamtrapuser[i], config->spamtrapdom[i]);
		}

	}

}

void cleanup(void) {

	char *p;
	int i;

	for(i=0;i<MAX_HTTPPORTS+MAX_SOCKSPORTS+1;i++)
		free(config->port[i]);
	free(config->port);
	if (config->file->http) free(config->file->http);
	if (config->file->smtp) free(config->file->smtp);
	if (config->file->sock) free(config->file->sock);
	if (config->file->rdns) free(config->file->rdns);
	if (config->file->trap) free(config->file->trap);
	if (config->file->localdb) free(config->file->localdb);

	free(config->file);

	if (config->testhost) {
		for (i=0;config->testhost[i];i++) free(config->testhost[i]);
		free(config->testhost);
	}

	if (config->dbhome) free(config->dbhome);

	if (config->config) free(config->config);

	if (config->listenaddr) free(config->listenaddr);

	if (config->spamtrapuser) {
		for (i=0;config->spamtrapuser[i];i++) free(config->spamtrapuser[i]);
		free(config->spamtrapuser);
	}

	free(config->exitval);
	free(config);

}
