/* Debug Header */
#if DEBUG
# define IFDEBUG(x) do { if (debug) x; }while(0)
# define IFDEBUG5(x) do { if (debug) x; }while(0)
#else
# define IFDEBUG(x)
# define IFDEBUG5(x)
#endif
/* End Debug Header */


#if HAVE_CONFIG_H
# include <config.h>
#endif

/* network.c
 * 
 *  a collection of general network functions.
 */


#include <strings.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netdb.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#include "dbconfig.h"
#include "extern.h"
#include "defines.h"

/* start_listener():
 *
 * needs: open socket bound and listening to port.
 *
 * returns: doesn't .... while(1) ....
 *
 * purpose: accept incoming connections, spawning new thread to handle the incoming connection.
 *
 * on error, will call exit() - if thread start error, will close the connection and try again.
 */
void start_listener(int sd) {

	int csock=0;
	int clientlen;
	struct hostent *host;
	struct sockaddr_in clientaddr;
	char *ipaddr;
	char *hostname;
	int err=0;

	IFDEBUG(print_debug(1, "Waiting for connections on Port %i.\n", config->listenport));
	
	while(1) {


		clientlen = sizeof(clientaddr);

		csock = accept(sd, (struct sockaddr *) &clientaddr, &clientlen);

		IFDEBUG(print_debug(2, "Accepted incoming connection - processing...\n"));

		ipaddr=(char *)strdup(inet_ntoa(clientaddr.sin_addr));
		host=gethostbyaddr(&clientaddr.sin_addr, sizeof(clientaddr.sin_addr), PF_INET);
		if (!host) {
			herror("Error");
			hostname=(char *)calloc(sizeof(char), strlen(ipaddr)+3);
			sprintf(hostname, "[%s]", ipaddr);
		} else {
			hostname=(char *)strdup(host->h_name);
		}
		IFDEBUG(print_debug(3, "Connection from remote host: %s [%s]\n", hostname, ipaddr));
		if (sd<0) {
			perror("Fatal: cannot accept connection");
			cleanup();
			exit(1);
		}
		fflush(stderr);
		fflush(stdout);
		if (start_client_thread(csock, ipaddr, hostname) < 0) {
			fprintf(stderr, "Unable to start new thread for connection - dropping\n");
			close(csock);
		}
	}
}


/* setup_port():
 *
 * needs: nothing
 *
 * returns:i (int)  open network socket listening to port configured in config file.
 *
 * will call exit() if unable to bind to port.
 */
int setup_port(void) {

	int connection=0;
	int sockfd=0;

	struct hostent *host;
	struct sockaddr_in servaddr;

	sockfd=socket(PF_INET, SOCK_STREAM, 0);

	if (config->listenaddr) {
		host=gethostbyname(config->listenaddr);
	} else {
		host=gethostbyname("0.0.0.0");
	}

	servaddr.sin_port=htons(config->listenport);
	//memcpy((char *) &servaddr.sin_addr.s_addr, host?host->h_addr_list[0]:INADDR_ANY, host?host->h_length:sizeof(INADDR_ANY));
	if (host) {
		servaddr.sin_family=host->h_addrtype;
		memcpy((char *) &servaddr.sin_addr.s_addr, host->h_addr_list[0], host->h_length);
	} else {
		servaddr.sin_family=PF_INET;
		servaddr.sin_addr.s_addr = htonl (INADDR_ANY);
	}

	connection=bind(sockfd, (struct sockaddr *)&servaddr, sizeof(servaddr));

	if (connection) {
		perror("Fatal: cannot bind to port");
		cleanup();
		exit(1);
	}

	listen(sockfd,5);
	return sockfd;

}

/* net_connect(): 
 * accepts							returns
 * hostname (char *)
 * port number (int)
 *											socket file descriptor number
 *
 * this function is just to open a socket
 * to a remote host given the hostname and
 * port number.  It returns the fd number of
 * the open socket.
 */
int net_connect(const char *hostname, const int port) {

	int sockfd, connection, ret;
	int error=0;
	struct hostent *ipaddr;
	struct sockaddr_in servaddr;
	struct timeval tv; 
	fd_set fdw, fdr;
	int i;

	if (verbose) printf("Connecting to %s:%i\n", hostname, port);
	IFDEBUG(print_debug(4, "net_connect(): hostname=%s, port=%i\n", hostname, port));
	ipaddr=gethostbyname(hostname);
	if (!ipaddr) {
		for(i=0;i<5;i++) {
			ipaddr=gethostbyname(hostname);
			if (ipaddr) break;
			sleep(1);
		}
		herror("Fatal Error");
		return(0);
	}
	sockfd=socket(PF_INET, SOCK_STREAM, 0);
	fcntl(sockfd, F_SETFL, O_NONBLOCK);

	servaddr.sin_family=ipaddr->h_addrtype;
	servaddr.sin_port=htons(port);
	memcpy((char *) &servaddr.sin_addr.s_addr, ipaddr->h_addr_list[0], ipaddr->h_length);
	connection=connect(sockfd, (struct sockaddr *)&servaddr, sizeof(servaddr));

	tv.tv_sec = 30; 
	tv.tv_usec = 0; 

	if (connection) {
		switch (errno) {
			case EAGAIN: /* we can handle this error */
			case EINPROGRESS: 

				FD_ZERO(&fdw);
				FD_SET(sockfd, &fdw);
				fdr = fdw ;
	
				tv.tv_sec = NET_TIMEOUT; 
				tv.tv_usec = 0; 

				if (select(sockfd + 1, &fdr, &fdw, NULL, &tv ) != 0) {
					/* even if socket is writeable, may be an error, need explicit check */
					if (FD_ISSET(sockfd, &fdr) || FD_ISSET(sockfd, &fdw)) {
						int len = sizeof(error);
						if (getsockopt(sockfd, SOL_SOCKET, SO_ERROR, (char *)&error, &len) == 0) {
							/* solaris getsockopt() will return -1 if an
 								 * error has occurred  otherwise we must
 								 * explicitly check "error"
							 */
							if (error == 0) { /* no errors on connect */
								if (verbose) printf("Connection Established, testing....\n");

								/* set 30 second tiemouts on read/writes */
								tv.tv_sec = NET_DTIMEOUT; 
								tv.tv_usec = 0; 

								setsockopt(sockfd, SOL_SOCKET, SO_RCVTIMEO, &tv, sizeof(tv)); 
								setsockopt(sockfd, SOL_SOCKET, SO_SNDTIMEO, &tv, sizeof(tv)); 
														
								break; /* let's get out of here ... */
							}
						}
					}
				}
				/* Fall through to default */
			default: 
				if (verbose) fprintf(stderr, "Timed Out connection.\n");
				close(sockfd); /* close socket, and (maybe) try again */
				sockfd=0;
				break;
		}
	}
	return sockfd;
}

	 
/* net_read():
 *
 * Needs: Open network socket
 *
 * Returns: Arbatry string read from network socket, NULL on error
 */

char *net_read(int sockfd) {	

	int num, r;
	char *str;

	fd_set rdfs;
	struct timeval tv;
	int ret, timeout, i;
	char *buf, *ptr;
	
	unsigned char t[1];


	/* clear string contents and initialise all to null. Good and safe practice */	
	str=malloc(NET_BUFFER+1);	
	memset(str, '\0', NET_BUFFER+1);

	FD_ZERO(&rdfs);
	FD_SET(sockfd, &rdfs);

	tv.tv_sec = NET_RTIMEOUT;
	tv.tv_usec = 0;

	num=0;
	ret=select((sockfd+1), &rdfs, NULL, NULL, &tv);
	if (ret) {
		num=recv(sockfd, str, NET_BUFFER-1,0);

		if (!num && !ret) {
			perror("An Error Occurred");
		}
	}

	if (!num)	{
		free(str);
		return NULL; /* safely return something worth while. */	
	}


	/* yuck! blackslashes are needed for some preprocessors. */
	IFDEBUG5({ \
		printf("Debug[5]: net_read() "); \
		for(r=0;r<num;r++) { \
			t[0]=*(str+r); \
			printf("%hhi ", t[0]); \
		} \
		printf("["); \
		for(r=0;r<num;r++) { \
			t[0]=*(str+r)>31?127>*(str+r)?*(str+r):'.':'.'; \
			printf("%c", t[0]); \
		} \
		printf("]\n"); \
	});

	/* return the damn stuff.  */	
	return str;
}

/* net_send():
 *
 * needs: open socket
 *        string to sent to socket
 *
 * returns: -1 on error or 0 for success
 */

int net_send(int sockfd, char *str) {
	
	int r;

	IFDEBUG(print_debug(4, "Sending: %s", str));
	if (!send(sockfd, str, strlen(str), 0)) {
		perror("Error sending to network");
		return -1;
	}
	return 0;

}

/* net_nsend():
 *
 * needs: open socket
 *        string/data to send to socket
 *        length of data to send
 *
 * returns: -1 on error or 0 for success
 */
int net_nsend(int sockfd, char *str, int len) {
	
	int r;
	unsigned char t[1];

	/* yuck! blackslashes are needed for some preprocessors. */
	IFDEBUG5({ \
		printf("Debug[5]: net_nsend() "); \
		for(r=0;r<len;r++) { \
			t[0]=*(str+r); \
			printf("%hhi ", t[0]); \
		} \
		printf("\n"); \
	});

	if (!send(sockfd, str, len, 0)) {
		perror("Error sending to network");
		return -1;
	}
	return 0;

}

/* getip():
 *
 * needs: (char *) hostname/ipaddress
 *
 * returns: (char *) ipaddress, or passed string on error.
 */
char *getip(const char *hostname) {

	int sockfd, connection;
	struct hostent *ipaddr;
	struct sockaddr_in servaddr;
	int i;

	ipaddr=gethostbyname(hostname);
	if (!ipaddr) {
		for(i=0;i<5;i++) {
			ipaddr=gethostbyname(hostname);
			if (ipaddr) break;
			sleep(1);
		}
		if (!ipaddr) {
			fprintf(stderr, "Error unable to reverse map hostname %s", hostname);
			herror("in DNS");
			return((char *)hostname);
		}
	}
	return ipaddr->h_addr_list[0];
}

